import BigNumber from 'bignumber.js';
import { Abi, Address } from '../../utils/basicTypes';
import { EthCallTx, EthSendTx, SendOptions, Web3TransactionResult } from '../../utils/transactions';
import EthLibAdapter, { Contract } from '../EthLibAdapter';
export declare function toTxResult(promiEvent: any, sendOptions?: SendOptions): Promise<Web3TransactionResult>;
export interface Web3AdapterConfig {
    web3: any;
}
declare class Web3Adapter extends EthLibAdapter {
    web3: any;
    /**
     * Creates an instance of Web3Adapter
     *
     * @param options - Web3Adapter configuration
     * @returns The Web3Adapter instance
     */
    constructor({ web3 }: Web3AdapterConfig);
    /**
     * Returns the current provider
     *
     * @returns The current provider
     */
    getProvider(): any;
    /**
     * Sends a network request via JSON-RPC.
     *
     * @param method - JSON-RPC method
     * @param params - Params
     * @returns The request response
     */
    providerSend(method: string, params: any[]): Promise<any>;
    /**
     * Signs data using a specific account.
     *
     * @param message - Data to sign
     * @param ownerAccount - Address to sign the data with
     * @returns The signature
     */
    signMessage(message: string, ownerAccount: Address): Promise<string>;
    /**
     * Returns the current network ID.
     *
     * @returns The network ID
     */
    getNetworkId(): Promise<number>;
    /**
     * Returns the default account used as the default "from" property.
     *
     * @returns The default account address
     */
    getAccount(): Promise<Address>;
    /**
     * Returns the balance of an address.
     *
     * @param address - The desired address
     * @returns The balance of the address
     */
    getBalance(address: Address): Promise<BigNumber>;
    /**
     * Returns the keccak256 hash of the data.
     *
     * @param data - Desired data
     * @returns The keccak256 of the data
     */
    keccak256(data: string): string;
    /**
     * Encodes a function parameters based on its JSON interface object.
     *
     * @param types - An array with the types or a JSON interface of a function
     * @param values - The parameters to encode
     * @returns The ABI encoded parameters
     */
    abiEncode(types: string[], values: any[]): string;
    /**
     * Decodes ABI encoded parameters to is JavaScript types.
     *
     * @param types - An array with the types or a JSON interface outputs array
     * @param data - The ABI byte code to decode
     * @returns The ABI encoded parameters
     */
    abiDecode(types: string[], data: string): any[];
    /**
     * Deterministically returns the address where a contract will be deployed.
     *
     * @param deployer - Account that deploys the contract
     * @param salt - Salt
     * @param initCode - Code to be deployed
     * @returns The address where the contract will be deployed
     */
    calcCreate2Address(deployer: Address, salt: string, initCode: string): string;
    /**
     * Returns the code at a specific address.
     *
     * @param address - The desired address
     * @returns The code of the contract
     */
    getCode(address: Address): Promise<string>;
    /**
     * Returns a block matching the block number or block hash.
     *
     * @param blockHashOrBlockNumber - The block number or block hash
     * @returns The block object
     */
    getBlock(blockHashOrBlockNumber: string | number): Promise<{
        [property: string]: any;
    }>;
    /**
     * Returns an instance of a contract.
     *
     * @param abi - ABI of the desired contract
     * @param address - Contract address
     * @returns The contract instance
     */
    getContract(abi: Abi, address: Address): Contract;
    /**
     * Returns the revert reason when a call fails.
     *
     * @param tx - Transaction to execute
     * @param block - Block number
     * @returns The revert data when the call fails
     */
    getCallRevertData(tx: EthCallTx, block: string | number): Promise<string>;
    /**
     * Sends a transaction to the network.
     *
     * @param tx - Transaction to send
     * @returns The transaction response
     */
    ethSendTransaction(tx: EthSendTx): Promise<Web3TransactionResult>;
    /**
     * Formats transaction result depending on the current provider.
     *
     * @param txHash - Transaction hash
     * @param tx - Transaction response
     * @returns The formatted transaction response
     */
    toSafeRelayTxResult(txHash: string, tx: Record<string, any>): Promise<Web3TransactionResult>;
    toRocksideRelayTxResult(tx: Record<string, any>): Promise<Web3TransactionResult>;
}
export default Web3Adapter;
